#include <libsgspin.h>
#include "graph.h"

Node::Node(StateBase *st, short autom_num, int autom_trans_id, int ltl_trans_id)
{
	lowlink = 0;
	number = 0;
	LastChild = NULL;
	Children = NULL;
	state = st;
	parent = 0;
	m_flags = NONE;
	component_number = 0;
	packed_state = new PackedState;
	packed_state->autom_trans_id = autom_trans_id;
	packed_state->ltl_trans_id = ltl_trans_id;
	packed_state->autom_number = autom_num;
	StateHash::pack_state_hash_key(st, &(packed_state->hash_key) );
}

Node::~Node()
{
	if(state)
		delete state;
	delete[] packed_state->hash_key;
	delete packed_state;
}

/*
	PDS: An awful lot of allocation for comparison.
*/
bool
Node::operator==(Node &otherNode)
{
	bool eq, delother, delcurr;
	StateBase *other, *curr;
	uchar *otherbytes, *currbytes;
	int othersize, currsize;

	eq = memcmp(packed_state->hash_key,
		otherNode.packed_state->hash_key,
		StateHash::HASH_KEY_BYTE_SIZE) == 0;
	hash_comparison++;
	if(eq){
		other = otherNode.GetState();
		curr = state;
		delother = false;
		delcurr = false;
		if(other == 0 && detect_collisions){
			other = otherNode.GenerateNodeState();
			delother = true;
			state_restoring++;
		}
		if(curr == 0 && detect_collisions){
			curr = GenerateNodeState();
			delcurr = true;
			state_restoring++;
		}
		if(curr && other){
			currsize = curr->get_state_byte_array(&currbytes);
			othersize = other->get_state_byte_array(&otherbytes);
			eq = currsize == othersize
				&& memcmp(currbytes,otherbytes,currsize) == 0;
			delete[] currbytes;
			delete[] otherbytes;
			state_comparison++;
			/*
				PDS: Code to compare other's and
				curr's proc vectors, never claims,
				and global vars was commented
				out.
			*/
			if(!eq)
				collisions_count++;
		}
		if(delother)
			delete other;
		if(delcurr)
			delete curr;
	}
	return eq;
}

void
Node::SetLowLink(int ll)
{
	lowlink = ll;
}

void
Node::SetNumber(int num)
{
	number = num;
}

int
Node::GetLowLink()
{
	return lowlink;
}

int
Node::GetNumber()
{
	return number;
}

StateBase*
Node::GetState()
{
	return state;
}

PackedState*
Node::GetPackedState()
{
	return packed_state;
}

Node*
Node::GetParent()
{
	return parent;
}

void
Node::ResetLastChild()
{
	LastChild = NULL;
}

Node*
Node::GetLastChild()
{
	if(LastChild != NULL)
		return LastChild->node;
	return NULL;
}

Node*
Node::GetNextChild()
{
	Node* r;

	if(LastChild == NULL)
		LastChild = Children;
	else
		LastChild = LastChild->next;
	r = GetLastChild();
	return r;
}

bool
Node::LastChildIsFirst()
{
	return LastChild == Children;
}

void
Node::SetCurrNodeElement(NodeElement* el)
{
	LastChild = el;
}

NodeElement*
Node::GetCurrNodeElement()
{
	return LastChild;
}

bool
Node::InScenario()
{ 
	return ((m_flags & IN_SCENARIO) == IN_SCENARIO);
}

void
Node::SetInScenario(bool in_scenario)
{
	if(in_scenario)
		m_flags |= IN_SCENARIO;
	else
		m_flags &= (~IN_SCENARIO);
}

bool
Node::InScenarioNotInRS()
{ 
	return ((m_flags & R_SCENARIO) == R_SCENARIO); 
}

void
Node::SetInScenarioNotInRS(bool r_scenario)
{
	if(r_scenario)
		m_flags |= R_SCENARIO;
	else
		m_flags &= (~R_SCENARIO);
}

bool
Node::IsFixed()
{ 
	return ((m_flags & FIXED) == FIXED); 
}

void 
Node::SetFixed(bool fixed)
{
	if(fixed)
		m_flags |= FIXED;
	else
		m_flags &= (~FIXED);
}

bool
Node::IsPassed()
{ 
	return ((m_flags & PASSED) == PASSED); 
}

void
Node::SetPassed(bool passed)
{
	if(passed)
		m_flags |= PASSED;
	else
		m_flags &= (~PASSED);
}

void 
Node::SetFlags(uchar flags, bool enable)
{
	if(enable)
		m_flags |= flags;
	else
		m_flags &= (~flags);
}

int
Node::GetComponentNumber()
{ 
	return component_number; 
}

void
Node::SetComponentNumber(int number)
{ 
	component_number = number; 
}

void
Node::SetBorderTransStart(bool border)
{
	if(border)
		m_flags |= TRANS_TO_BORDER;
	else
		m_flags &= (~TRANS_TO_BORDER);
}

bool
Node::HasBorderChild()
{ 
	return ((m_flags & TRANS_TO_BORDER) == TRANS_TO_BORDER); 
}

void
Node::RemoveDuplicatedChildren()
{
	NodeElement *child, *del, *prev_child;
	
	prev_child = 0;
	LastChild = NULL;
	child = Children;
	while(child){
		if(child->node->IsPassed()){
			del = child;
			child = child->next;
			prev_child->next = child;
			delete del;
		}
		else{
			child->node->SetPassed(true);
			prev_child = child;
			child = child->next;
		}
	}

	child = Children;
	for(; child; child = child->next)
		child->node->SetPassed(false);
}

StateBase*
Node::GenerateNodeState()
{
	Node *n, *n_parent, *n_curr;
	int n_ind, state_count;
	NodeStack stack;
	StateBase *curr_state, *st;
	CAutomBase *never, *autom;
	PackedState *pack_state;
	NodeElement *last_child;
	unsigned int st_size;
	list<StateBase*> ltl_ret_states, ret_states;

	/*
		Reconstruct the DFS stack as it was when
		this node was first found using saved parent
		pointers.
	*/
	for(n=this; n->GetParent(); n=n->GetParent()){
		stack.Push(n);
		reconstructing_stack_summ++;
	}
	/*
		Reconstruct this node's State using saved
		transition information.  (The initial node
		n contains a non-zero State.)
	*/
	curr_state = (StateBase*)n->GetState()->clone();
	while(!stack.Empty()){
		n = stack.Pop();
		n_parent = n->GetParent();
		pack_state = n->GetPackedState();
		
		never = curr_state->get_never_claim();
		if(never)
			never->new_state(curr_state, pack_state->ltl_trans_id);
		
		/* determine the index of n in n_parent.Children[] */
		n_ind = 0;
		last_child = n_parent->GetCurrNodeElement();
		n_parent->ResetLastChild();
		while((n_curr = n_parent->GetNextChild()) != 0){
			if (n_curr == n)
				break;
			else if(n_curr->GetPackedState()->ltl_trans_id == pack_state->ltl_trans_id
			&& n_curr->GetPackedState()->autom_trans_id == pack_state->autom_trans_id
			&& n_curr->GetPackedState()->autom_number == pack_state->autom_number)
				n_ind++;
		}
		n_parent->SetCurrNodeElement(last_child);

		/*
			PDS: Unobfuscate this code and examine code
			that might have side-effects on curr_state
			(like get_proc()->new_state(), etc.)
		*/
		if (never && never->is_inside_atomic()){
			state_count = 0;
			GenerateAtomicStates(-1, curr_state, ltl_ret_states);
			while(!ltl_ret_states.empty()){
				st = ltl_ret_states.back(); 
				ltl_ret_states.pop_back();
				autom = st->get_proc(pack_state->autom_number);
				autom->new_state(st, pack_state->autom_trans_id);
				if(autom->is_inside_atomic()){
					GenerateAtomicStates(pack_state->autom_number, curr_state, ret_states);
					state_count += ret_states.size();
					if(n_ind < state_count){
						n_ind = ret_states.size() - (state_count - n_ind);
						while(n_ind > 0){
							delete ret_states.back();
							ret_states.pop_back();
							n_ind--;
						}
						curr_state = ret_states.back();
						ret_states.pop_back();
						while(!ret_states.empty()){
							delete ret_states.back();
							ret_states.pop_back();
						}
						break;
					}
				}
				delete st;
			}
			while (!ltl_ret_states.empty()){
				delete ltl_ret_states.back();
				ltl_ret_states.pop_back();
			}
		}
		else{
			autom = curr_state->get_proc(pack_state->autom_number);
			autom->new_state(curr_state, pack_state->autom_trans_id);
			if(autom->is_inside_atomic()){
				GenerateAtomicStates(pack_state->autom_number, curr_state, ret_states);
				while(n_ind > 0){
					delete ret_states.back();
					ret_states.pop_back();
					n_ind--;
				}
				curr_state = ret_states.back();
				ret_states.pop_back();
				while(!ret_states.empty()){
					delete ret_states.back();
					ret_states.pop_back();
				}
	
			}
		}
	}
	return curr_state;
}

void
Node::GenerateAtomicStates(int procnum, StateBase *state, list<StateBase*> &states)
{
	list<StateBase*> curr_states;
	CAutomBase *proc, *next_proc;
	int else_trans_id, state_num, trans_id, new_trans_id;
	bool enable_trans;
	StateBase* next_state;

	curr_states.push_front(state);
	while (!curr_states.empty()) {
		state = curr_states.front(); 
		curr_states.pop_front();
		proc = state->get_proc(procnum);
		proc->set_first_skip_trans(false);
		if(procnum < 0)
			proc = state->get_never_claim();
		else
			proc = state->get_proc(procnum);
		proc->set_moved(false);
		proc->set_else_passed(false);
		else_trans_id = -1;
		state_num = proc->get_curr_state();
		enable_trans = false;
		for(	trans_id = proc->get_first_trans_id(state_num);
			trans_id != 0;
			trans_id = proc->get_next_trans_id(trans_id))
		{
			if(!proc->is_trans_enable(state,trans_id)){
				if(proc->get_next_trans_id(trans_id) == 0
				&& proc->is_else_passed()
				&& !proc->is_moved()){
					trans_id = else_trans_id;
					enable_trans = true;
				}
				else
					continue;
			}
			else{
				if(proc->is_else_passed()
				&& proc->get_next_trans_id(trans_id) != 0){
					else_trans_id = trans_id;
					continue;
				}
				enable_trans = true;
			}
			if(enable_trans){
				next_state = (StateBase*)state->clone();
				if(procnum < 0)
					next_proc = next_state->get_never_claim();
				else
					next_proc = next_state->get_proc(procnum);
				next_proc->set_moved(false);
				next_proc->new_state(next_state, trans_id);
				proc->set_moved(true);
				new_trans_id = next_proc->get_first_trans_id(next_proc->get_curr_state());
				if(next_proc->is_inside_atomic())
					curr_states.push_front(next_state);
				else
					states.push_front(next_state);
			}
			if(trans_id == else_trans_id)
				break;
		}
		if(!enable_trans)
			states.push_back(state);
		else
			delete state;
	}
}
	
void
Node::GenerateChildren()
{
	CAutomBase *never, *never_new, *proc, *proc_new;
	int orig_state, trans_id, else_trans_id, pr_trans_id, proc_cnt, procnum;
	StateBase *NextState, *st;
	list<StateBase*> ret_states;
	
	/*
		PDS: Lots of duplicate code for never and procs and
		between GenerateChildren and AutomsProcesing.
		Simplify.
	*/
	/*
		else_trans_id will contain id of "else" statement (or
		atomic statement that contains "else")
	*/
	never = state->get_never_claim();
	if(never != 0){	/* never claim moves first */
		orig_state = never->get_curr_state();
		trans_id = never->get_first_trans_id(orig_state);
		else_trans_id = -1;
		never->set_moved(false);
		never->set_else_passed(false);

		for(trans_id = never->get_first_trans_id(orig_state); trans_id != 0; trans_id = never->get_next_trans_id(trans_id)){
			if(!never->is_trans_enable(state, trans_id)){
				never->set_curr_state(orig_state);
				if(never->get_next_trans_id(trans_id) == 0
				&& never->is_else_passed()
				&& !never->is_moved())
					trans_id = else_trans_id;
				else
					continue;
			}
			else{
				never->set_curr_state(orig_state);
				if(never->is_else_passed()
				&& never->get_next_trans_id(trans_id) != 0){
					else_trans_id = trans_id;
					continue;
				}
			}

			NextState = (StateBase*)state->clone();
			never_new = NextState->get_never_claim();
			never_new->new_state(NextState, trans_id);

			never->set_moved(true);

			if(never_new->is_inside_atomic()){
				GenerateAtomicStates(-1, NextState, ret_states);
				while(!ret_states.empty()){
					st = ret_states.front();
					ret_states.pop_front();
					AutomsProcesing(state, st, trans_id);
					delete st;
				}
			}
			else{
				AutomsProcesing(state, NextState, trans_id);
				delete NextState;
			}

			/*
				"else" transition should always be last.
			*/
			if(trans_id == else_trans_id)
				break;
		}
	}
	/*
		PDS: The comment "never claim moves first" suggests
		that processes get to move too.  Does the "else"
		belong here?
	*/
	else{
		proc_cnt = state->proc_count();
		for (procnum = 0; procnum < proc_cnt; procnum++){
			proc = state->get_proc(procnum);

			if(proc != 0){
				else_trans_id = -1;
				orig_state = proc->get_curr_state();
				pr_trans_id = proc->get_first_trans_id(orig_state);	/* PDS: Why? */
				proc->set_moved(false);
				proc->set_else_passed(false);

				for (pr_trans_id = proc->get_first_trans_id(proc->get_curr_state()); pr_trans_id != 0; pr_trans_id = proc->get_next_trans_id(pr_trans_id)){
					if(!proc->is_trans_enable(state, pr_trans_id)){
						proc->set_curr_state(orig_state);
						if(proc->get_next_trans_id(pr_trans_id) == 0
						&& proc->is_else_passed()
						&& !proc->is_moved())
							pr_trans_id = else_trans_id;
						else
							continue;
					}
					else{
						proc->set_curr_state(orig_state);
						if (proc->is_else_passed()
						&& proc->get_next_trans_id(pr_trans_id) != 0)
						{
							else_trans_id = pr_trans_id;
							continue;
						}
					}

					NextState = (StateBase*)state->clone();
					proc_new = NextState->get_proc(procnum);
					proc_new->new_state(NextState, pr_trans_id);

					proc->set_moved(true);

					if(proc_new->is_inside_atomic()){
						GenerateAtomicStates(procnum, NextState, ret_states);
						while (!ret_states.empty()){
							st = ret_states.front();
							ret_states.pop_front();
							GenerateChild(st, procnum, pr_trans_id, 0);
						}
					}
					else{
						GenerateChild(NextState, procnum, pr_trans_id, 0);
					}

					/*
						"else" transition should always be last.
					*/
					if(pr_trans_id == else_trans_id)
						break;
				}
			}
		}
	}
}

void
Node::AutomsProcesing(StateBase *state, StateBase *NextState_never, int ltl_trans_id)
{
	int proc_cnt, procnum, else_trans_id, orig_state, pr_trans_id;
	StateBase *NextState, *st;
	CAutomBase *proc, *proc_new;
	list<StateBase*> ret_states;
	
	proc_cnt = state->proc_count();
	// One of the processes moves
	for(procnum = 0; procnum < proc_cnt; procnum++){
		proc = state->get_proc(procnum);
		if(proc != 0){
			else_trans_id = -1;
			orig_state = proc->get_curr_state();
			pr_trans_id = proc->get_first_trans_id(orig_state);
			proc->set_moved(false);
			proc->set_else_passed(false);

			for(pr_trans_id = proc->get_first_trans_id(orig_state); pr_trans_id != 0; pr_trans_id = proc->get_next_trans_id(pr_trans_id)){
				if(!proc->is_trans_enable(state, pr_trans_id)){
					proc->set_curr_state(orig_state);
					if(proc->get_next_trans_id(pr_trans_id) == 0
					&& proc->is_else_passed()
					&& !proc->is_moved())
						pr_trans_id = else_trans_id;
					else
						continue;
				}
				else{
					proc->set_curr_state(orig_state);
					if(proc->is_else_passed()
					&& proc->get_next_trans_id(pr_trans_id) != 0)
					{
						else_trans_id = pr_trans_id;
						continue;
					}
				}

				NextState = (StateBase*)NextState_never->clone();
				proc_new = NextState->get_proc(procnum);
				proc_new->set_moved(false);
				proc_new->new_state(NextState, pr_trans_id);

				proc->set_moved(true);

				if(proc_new->is_inside_atomic()){
					GenerateAtomicStates(procnum, NextState, ret_states);
					while (!ret_states.empty()){
						st = ret_states.front();
						ret_states.pop_front();
						GenerateChild(st, procnum, pr_trans_id, ltl_trans_id);
/*						AddChild(...); */
					}
				}
				else{
					GenerateChild(NextState, procnum, pr_trans_id, ltl_trans_id);
/*						AddChild(...); */
				}
				if(pr_trans_id == else_trans_id)
					break;
			}
		}
	}
}

void
Node::AddChild(Node *node)
{	
	NodeElement *e;
	
	e = new NodeElement;
	if(node->parent == 0)
		node->parent = this;
	e->node = node;
	e->next = Children;
	Children = e;
}

Node*
Node::GenerateChild(int procnum, int pr_trans_id, int trans_id)
{
	StateBase *NextState;
	CAutomBase *never, *proc;
	Node *NextChild;
	NodeHashSet::iterator PreviouslyFound;
	
	NextState = (StateBase*)state->clone();
	never = NextState->get_never_claim();
	if(never != 0){
		never->new_state(NextState, trans_id);
		if(never->is_stop_found_first()){
			ForceAccepting(true);
			ForceFinal(true);
		}
	}
	proc = NextState->get_proc(procnum);
	proc->new_state(NextState, pr_trans_id);

	NextChild = new Node(NextState, procnum, pr_trans_id, trans_id);
	PreviouslyFound = Nodes.find(NextChild);

	dfs_edges_count++;

	if(PreviouslyFound != Nodes.end()){
		delete NextChild;
		return *PreviouslyFound;
	}
	else{
		unsigned int st_size = NextState->calc_size();
		if(st_size > StateBase::max_state_size)
			StateBase::max_state_size = st_size;
		Nodes.insert(NextChild);
		NodeList.push_back(NextChild);
		return NextChild;
	}
}

void
Node::GenerateChild(StateBase *NextState, short autom_num, int autom_trans_id, int ltl_trans_id)
{
	Node* NextChild;
	NodeHashSet::iterator PreviouslyFound;
	CAutomBase *never;
	unsigned int st_size;
	long depth;

	NextChild = new Node(NextState, autom_num, autom_trans_id, ltl_trans_id);
	PreviouslyFound = Nodes.find(NextChild);

	dfs_edges_count++;

	never = NextState->get_never_claim();
	if(never != 0){
		if(never->is_stop_found_first()){
			ForceAccepting( true );
			ForceFinal( true );
		}
	}

	if(PreviouslyFound != Nodes.end()){
		delete NextChild;
		AddChild(*PreviouslyFound);
	}
	else{
		st_size = NextState->calc_size();
		depth   = 0;
		if(st_size > StateBase::max_state_size)
			StateBase::max_state_size = st_size;

		Nodes.insert(NextChild);
		NodeList.push_back(NextChild);
		AddChild(NextChild);

		while(NextChild != 0){
			NextChild = NextChild->parent;
			depth++;
		}

		if(depth > max_depth)
			max_depth = depth;
	}

	if(do_lefty_output)
		printf("\n\n");
}

bool
Node::HasSelfLoop()
{
	NodeElement *Child;
	
	for(Child = Children; Child; Child = Child->next){
		if(Child->node == this)
			return true;
	}
	return false;
}

bool
Node::Unexplored()
{
	return number==0;
}

void
Node::ForceAccepting(bool accp)
{
	if(accp)
		m_flags |= FORCED_ACCEPTED;
	else
		m_flags &= (~FORCED_ACCEPTED);
}

void
Node::ForceFinal(bool final)
{
	if(final)
		m_flags |= FORCED_FINAL;
	else
		m_flags &= (~FORCED_FINAL);
}

bool
Node::IsAccepting()
{
	CAutomBase *never;
	int s;
	bool r;

	if((m_flags&FORCED_ACCEPTED) == FORCED_ACCEPTED)
		return true;
	if(state==0)
		return false;
	never = state->get_never_claim();
	if(never == 0)
		return false;
	/*accpstate[NEVER][state->never]; */
	s = never->get_curr_state();
	r = never->is_state_accepted(s);
	return r;
}

bool
Node::IsFinal()
{
	CAutomBase *never;
	int s;
	bool r;

	if((m_flags&FORCED_FINAL) == FORCED_FINAL)
		return true;
	if(state==0)
		return false;
	never = state->get_never_claim();
	if(never==0)
		return false;
	s = never->get_curr_state();
	r = never->is_state_final(s);
	return r;
}

void
Node::Print()
{
	if (do_lefty_output)
		printf("%d|%d", number, lowlink);
}

void
Node::PrintSubgraphDotFormat(FILE *fd)
{
	int n, l, st, s, en;
	const char *shape;
	NodeElement *e;

	if(do_lefty_output){
		n = GetNumber();
		l = GetLowLink();
		st = 0;	/* XXX state->statenmum[0] */
		s = 0;	/* XXX state->globals.s */
		shape = IsAccepting() ? "doublecircle" : "circle";
		
		fprintf(fd, "	st%d [shape=%s, label=\"%d (%d) st %d\\ns=%d\"];\n",
			n, shape, n, l, st, s);

		for(e = Children; e != NULL; e = e->next){
			en = e->node->GetNumber();
			if(en > n)
				e->node->PrintSubgraphDotFormat(fd);
			fprintf(fd, "	st%d -> st%d;\n", n, en);
		}
	}
}

void
Node::PackNode()
{
	CAutomBase *never;
	int st;

	if(supertrace && parent != 0 && state != 0){
		never = state->get_never_claim();
		if(never != 0){
			st = never->get_curr_state();
			if(never->is_state_final(st))
				ForceFinal(true);
			if(never->is_state_accepted(st))
				ForceAccepting(true);
		}
		delete state;
		state = 0;
	}
}

unsigned int
Node::CalcSize(bool involve_states)
{
	unsigned int size;
	NodeElement *ch;
	
	size = sizeof(*this) + sizeof(PackedState) + StateHash::HASH_KEY_BYTE_SIZE + sizeof(NodeElement);
	for(ch = Children; ch != 0; ch = ch->next)
		size += sizeof(NodeElement);
	if(involve_states && state != 0)
		size += state->calc_size();
	return size;
}

void
Node::LeftyInsertNode(FILE *fd, char *GraphTableName, bool DFS_output)
{
	unsigned long flags, format_flags;
	ostream *dbg_stream;
	long ind;
	StateBase *curr_state;
	const char *details;

	flags = Tracer::get_flags();
	dbg_stream = Tracer::get_curr_stream();
	ind = GetNumber();
	curr_state = state;
	ofstream &out_graph = DFS_output ? dfs_graph_file : graph_file;

	if(curr_state == 0)
		curr_state = GenerateNodeState();

	if(ascii_file){
if(0) cerr << "Node " << GetNumber() << endl;
		Tracer::init_tracing(&out_graph, Tracer::GRAPH_ATTR_FORMAT | Tracer::ASCII_OUTPUT);
		out_graph << ind << ": Node{ bool: accept=" << (IsAccepting()?1:0) << "; ";
		out_graph << "bool: final=" << (IsFinal()?1:0) << "; ";
		if(DFS_output)
			out_graph << "bool: in_scenario=" << (InScenario()?1:0) << "; ";
		Tracer::trace(curr_state);
		out_graph << "}";
		Tracer::finish_ASCII_block();
	}
	else {
		format_flags = Tracer::GRAPH_ATTR_FORMAT;
		if(bin32_file)
			format_flags |= Tracer::BIN32;
		Tracer::init_tracing(&out_graph, format_flags);

		Tracer::write_bin_identifier(Tracer::get_name_id("Node"), &out_graph);
		Tracer::write_1bytes((char)-3, &out_graph);	//type "NODE_START"
		Tracer::write_4bytes(ind, &out_graph);	//value - node id

		Tracer::write_bin_identifier(Tracer::get_name_id("accept"), &out_graph);
		Tracer::write_1bytes(Tracer::get_type_id("bool"), &out_graph);	//type - "bool"
		Tracer::write_1bytes((unsigned char)(IsAccepting()?1:0), &out_graph);	//value

		Tracer::write_bin_identifier(Tracer::get_name_id("final"), &out_graph);
		Tracer::write_1bytes(Tracer::get_type_id("bool"), &out_graph);	//type - "bool"
		Tracer::write_1bytes((unsigned char)(IsFinal()?1:0), &out_graph);	//value

		if (DFS_output){
			Tracer::write_bin_identifier(Tracer::get_name_id("in_scenario"), &out_graph);
			Tracer::write_1bytes(Tracer::get_type_id("bool"), &out_graph);	//type - "bool"
			Tracer::write_1bytes((unsigned char)(InScenario()?1:0), &out_graph);	//value
		}

		Tracer::trace(curr_state);

		Tracer::write_bin_identifier(Tracer::get_name_id("Node"), &out_graph);
		Tracer::write_1bytes((char)-2, &out_graph);	//type "ATTR_END"
		Tracer::write_bin_identifier(-3, &out_graph);
	}

	/*
		Set to make lefty input (minimized) string (State,
		Automata, Glob./Loc.  variables)
	*/
	if(0){
		Tracer::set_flags(Tracer::LEFTY_INPUT);
		Tracer::trace(Tracer::LEFTY_INPUT | Tracer::TRACE_STATE, state);
		details = Tracer::get_output_str().c_str();
	}
	else
		details = " ";

	if(do_lefty_output)
		fprintf(fd, "fa.%s.insertnode(fa.%s, null, null, 'st%d', [\"shape\" = \"%s\"; \"label\" = \"%d (%d)\\n%s\";], 1);\n",
			GraphTableName, GraphTableName, GetNumber(),
			IsAccepting()?"doublecircle":"circle",
			GetNumber(), GetLowLink(), details);

	/*Restore tracing flags*/
	Tracer::init_tracing(dbg_stream, flags);

	Tracer::trace((Tracer *)0,GetNumber(),"Node: ", "\n");

	/*
		print State info (State, Automata (without tables),
		Queues, Queues el., Glob./Loc.  variables)
	*/
	flags = Tracer::TRACE_STATE
		| Tracer::TRACE_AUTOM
		| Tracer::TRACE_QUEUE
		| Tracer::EXPAND_QUEUE
		| Tracer::TRACE_VAR;
	if(curr_state)
		Tracer::trace(flags, curr_state);

	if(state == 0)
		delete curr_state;
}

void
Node::LeftyInsertEdge(FILE *fd, Node *source, Node *sink, char *GraphTableName, bool insert_to_file, bool DFS_output) 
{
	if(insert_to_file){
		ofstream &out_graph = ((DFS_output)? dfs_graph_file: graph_file);

		if(ascii_file)
			out_graph << source->GetNumber() << " -> " << sink->GetNumber() << "\n";
		else
		{
			Tracer::write_bin_identifier(-255, &out_graph);
			Tracer::write_4bytes((long)(source->GetNumber()), &out_graph);	//start node id
			Tracer::write_4bytes((long)(sink->GetNumber()), &out_graph);	//end node id
		}
	}
	if(do_lefty_output)
		fprintf(fd, "fa.%s.insertedge(fa.%s, fa.%s.graph.nodes[fa.%s.graph.nodedict[\"st%d\"]], null, fa.%s.graph.nodes[fa.%s.graph.nodedict[\"st%d\"]], null, null, 1);\n",
		    GraphTableName, GraphTableName, GraphTableName,
		    GraphTableName, source->GetNumber(),
		    GraphTableName, GraphTableName, sink->GetNumber());
}

void
Node::LeftyUpdateNode(FILE *fd, char *GraphTableName)
{
	unsigned long flags;
	string str;
	
	if(do_lefty_output){
		flags = Tracer::get_flags();
		/*
			Set to make lefty input (minimized) string
			(State, Automata, Glob./Loc.  variables)
		*/
		Tracer::set_flags(Tracer::LEFTY_INPUT);
		Tracer::trace(Tracer::LEFTY_INPUT | Tracer::TRACE_STATE, state);
		string str = Tracer::get_output_str();
		fprintf(fd, "fa.%s.graph.nodes[fa.%s.graph.nodedict[\"st%d\"]].attr.label = \"%d (%d)\\n%s\";\n",
			GraphTableName, GraphTableName, GetNumber(), GetNumber(), GetLowLink(),
			str.c_str());
		Tracer::set_flags(flags);
	}
}

void
Node::LeftyInsertNodeStep(FILE *fd, char *GraphTableName)
{
	LeftyInsertNode(fd, GraphTableName);

	if(do_lefty_output)
		fprintf(fd, "command done\n");
}

void
Node::LeftyInsertEdgeStep(FILE *fd, Node *source, Node *sink, char *GraphTableName)
{
	LeftyInsertEdge(fd, source, sink, GraphTableName);

	if(do_lefty_output)
		fprintf(fd, "command done\n");
}

void
Node::LeftyUpdateNodeStep(FILE *fd, char *GraphTableName)
{
	LeftyUpdateNode(fd, GraphTableName);

	if(do_lefty_output)
		fprintf(fd, "command done\n");
}

void
Node::LeftyHighlight(FILE *fd, char *GraphTableName)
{
	if(do_lefty_output)
		fprintf(fd, "highlightnode(fa.%s, fa.%s.graph.nodes[fa.%s.graph.nodedict[\"st%d\"]]);",
			GraphTableName, GraphTableName, GraphTableName, GetNumber());
}

void
Node::LeftyHighlightLabel(FILE *fd, char *GraphTableName)
{
	if(do_lefty_output)
		fprintf(fd, "highlightlabel(fa.%s, fa.%s.graph.nodes[fa.%s.graph.nodedict[\"st%d\"]]);",
			GraphTableName, GraphTableName, GraphTableName, GetNumber());
}

void
Node::LeftyHighlightEdge(FILE *fd, Node *source, Node *sink, char *GraphTableName)
{
	if(do_lefty_output)
		fprintf(fd, "highlightedge(fa.%s, fa.%s.graph.nodes[fa.%s.graph.nodedict[\"st%d\"]], fa.%s.graph.nodes[fa.%s.graph.nodedict[\"st%d\"]]);",
			GraphTableName, GraphTableName, GraphTableName, 
			source->GetNumber(), GraphTableName, GraphTableName, sink->GetNumber());
}

void
Node::PrintLeftyProgram(FILE *fd, char *GraphTableName)
{
	NodeElement *e;

	LeftyInsertNodeStep(fd, GraphTableName);
	for(e = Children; e != NULL; e = e->next){
		if(e->node->GetNumber() > GetNumber())
			e->node->PrintLeftyProgram(fd, GraphTableName);
		LeftyInsertEdgeStep(fd, this, e->node, GraphTableName);
	}
}

void
Node::ProfileHashCollisions()	/* PDS: Dead code. */
{
	ofstream coll;
	__gnu_cxx::hash<Node *> H;
	int nidx, total, collision;
	NodeHashSet::iterator nd, other;

	coll.open("coll");
	cout << endl;
	nidx = 0;
	total = 0;
	for(nd = Nodes.begin(); nd != Nodes.end(); nd++){
		collision = 0;
		for(other = Nodes.begin(); other != Nodes.end(); other++){
			if(H(*nd) == H(*other))
				collision++;
		}
		if(collision > 1){
			total++;
			coll << "node " << (*nd)->GetNumber() << ": " << collision << endl;
		}
		cout << "node " << nidx << ": " << collision-1 << ", " << total << endl;
		nidx++;
	}
	cout << endl;
	coll.close();
}

GraphMetrics::GraphMetrics():
	acceptance_number(0), max_node_out_degree(0),
	max_not_final_out_degree(0), graph_depth(0),
	longest_scenario(0), shortest_scenario(0)
{
}

Graph::Graph()
{
	edgecount = 0;
}


void
Graph::Add(Node *node)
{
	if(!node->InScenario()){
		node->LeftyInsertNode(stdout, "scen");
		node->PackNode();
		GraphNodes.push_back(node);
		node->SetInScenario(true);
	}
}

void
Graph::Add(NodeStack &nodes)
{
	Node *node;
	
	nodes.ResetIterator();
	for(node = nodes.Next(); node != NULL; node = nodes.Next()){
		if(!node->InScenario()){
			node->LeftyInsertNode(stdout, "scen");
			node->PackNode();
			GraphNodes.push_back(node);
			node->SetInScenario(true);
		}
	}
}

void
Graph::PrintToFile()
{
	list<Node*>::iterator nodeiter;
	Node *node;
	NodeElement *Child;
	
	for(nodeiter = GraphNodes.begin(); nodeiter != GraphNodes.end(); nodeiter++){
		node = *nodeiter;
		if(node->InScenario()){
			node->RemoveDuplicatedChildren();
			for(Child = node->Children; Child != NULL; Child = Child->next){
				if(Child->node->InScenario()){
					edgecount++;
					node->LeftyInsertEdge(stdout, node, Child->node, "scen", true);
				}
			}
		}
	}
}

unsigned int
Graph::GetNodeCount()
{
	return GraphNodes.size();
}

unsigned int
Graph::CalcSize()
{
	return sizeof(*this) + GraphNodes.size()*sizeof(Node*);
}


GraphMetrics
Graph::CountGraphMetrics(Node *init_node)
{
	GraphMetrics res;
	int counter, length, acc_number, max_degree, max_nf_degree;
	int graph_depth, shortest, longest;
	Node *node, *child;
	NodeStack PathNodesStack;
	list<Node*>::iterator nodeiter;
	
	counter = 0;
	length = 0;
	acc_number = 0;
	max_degree = 0;
	max_nf_degree = 0;
	graph_depth = 0;
	shortest = 0;
	longest  = 0;

	if(init_node->InScenario()){
		for(nodeiter = GraphNodes.begin(); nodeiter != GraphNodes.end(); nodeiter++){
			node = *nodeiter;
			node->RemoveDuplicatedChildren();
			node->ResetLastChild();

			child = node->GetNextChild();
			counter = 0;
			// calc node degree
			for(; child; child = node->GetNextChild())
				if(child->InScenario())
					counter++;

			if(counter > max_degree)
				max_degree = counter;

			if(node->IsAccepting())
				acc_number++;
			else if(counter > max_nf_degree)
				max_nf_degree = counter;
		}

		shortest = GraphNodes.size();
		counter = 0;

		PathNodesStack.Push(init_node);

		init_node->SetFixed(true);

		//calc scenario length 
		while(!PathNodesStack.Empty()){
			node = PathNodesStack.Top();
			child = node->GetNextChild();

			if(child == 0){
				PathNodesStack.Pop();
				node->SetFixed(false);
			}
			else if(child->InScenario()){
				if(child->IsFinal()){
					length = PathNodesStack.GetLength() + 1;

					if(longest < length)
						longest = length;

					if(shortest > length)
						shortest = length;

					counter ++;
				}
				else if(!child->IsFixed())
				{
					PathNodesStack.Push(child);
					child->SetFixed(true);
				}
			}
		}

		if(longest == 0)
			shortest = 0;

		PathNodesStack.Push(init_node);

		init_node->SetFixed(true);

		//calc whole graph depth 
		while(!PathNodesStack.Empty()){
			node = PathNodesStack.Top();
			child = node->GetNextChild();

			if(child == 0){
				length = PathNodesStack.GetLength();
				if(length > graph_depth)
					graph_depth = length;

				PathNodesStack.Pop();
				node->SetFixed(false);
			}
			else if(child->InScenario()){
				if(!child->IsFixed())
				{
					PathNodesStack.Push(child);
					child->SetFixed(true);
				}
			}
		}
	}

	res.acceptance_number = acc_number;
	res.max_node_out_degree = max_degree;
	res.max_not_final_out_degree = max_nf_degree;
	res.graph_depth = graph_depth;
	res.longest_scenario = longest;
	res.shortest_scenario = shortest;

	return res;
}

/*
	Support for NodeHashSet and NodePointerHashSet.
*/

typedef  size_t   ub4;        /* unsigned 4-byte quantities */
typedef  unsigned char ub1;   /* unsigned 1-byte quantities */

#define hashsize(n) ((ub4)1<<(n))
#define hashmask(n) (hashsize(n)-1)

/*
mix -- mix 3 32-bit values reversibly.
For every delta with one or two bits set, and the deltas of all three
  high bits or all three low bits, whether the original value of a,b,c
  is almost all zero or is uniformly distributed,
* If mix() is run forward or backward, at least 32 bits in a,b,c
  have at least 1/4 probability of changing.
* If mix() is run forward, every bit of c will change between 1/3 and
  2/3 of the time.  (Well, 22/100 and 78/100 for some 2-bit deltas.)
mix() was built out of 36 single-cycle latency instructions in a
  structure that could supported 2x parallelism, like so:
      a -= b;
      a -= c; x = (c>>13);
      b -= c; a ^= x;
      b -= a; x = (a<<8);
      c -= a; b ^= x;
      c -= b; x = (b>>13);
      ...
*/
#define mix(a,b,c) \
{ \
	a -= b; a -= c; a ^= (c>>13); \
	b -= c; b -= a; b ^= (a<<8); \
	c -= a; c -= b; c ^= (b>>13); \
	a -= b; a -= c; a ^= (c>>12);  \
	b -= c; b -= a; b ^= (a<<16); \
	c -= a; c -= b; c ^= (b>>5); \
	a -= b; a -= c; a ^= (c>>3);  \
	b -= c; b -= a; b ^= (a<<10); \
	c -= a; c -= b; c ^= (b>>15); \
}

/*
      jenkins_hash() -- hash a variable-length key into a 32-bit value
      k       : the key (the unaligned variable-length array of bytes)
      len     : the length of the key, counting by bytes
      initval : can be any 4-byte value
      Returns a 32-bit value.  Every bit of the key affects every bit of
      the return value.  Every 1-bit and 2-bit delta achieves avalanche.
      About 6*len+35 instructions.

      The best hash table sizes are powers of 2.  There is no need to do
      mod a prime (mod is sooo slow!).  If you need less than 32 bits,
      use a bitmask.  For example, if you need only 10 bits, do
      h = (h & hashmask(10));
      In which case, the hash table should have hashsize(10) elements.

      If you are hashing n strings (ub1 **)k, do it like this:
      for (i=0, h=0; i<n; ++i) h = hash( k[i], len[i], h);

      By Bob Jenkins, 1996.  bob_jenkins@burtleburtle.net.  You may use this
      code any way you wish, private, educational, or commercial.  It's free.

      See http://burtleburtle.net/bob/hash/evahash.html
      Use for hash table lookup, or anything where one collision in 2^^32 is
      acceptable.  Do NOT use for cryptographic purposes.
*/

static size_t
jenkins_hash(register ub1* k, register ub4 length, register ub4 initval)
{
	register ub4 a,b,c,len;

	/* Set up the internal state */
	len = length;
	a = b = 0x9e3779b9;  /* the golden ratio; an arbitrary value */
	c = initval;         /* the previous hash value */

	/*---------------------------------------- handle most of the key */
	while (len >= 12) {
		a += (k[0] +((ub4)k[1]<<8) +((ub4)k[2]<<16) +((ub4)k[3]<<24));
		b += (k[4] +((ub4)k[5]<<8) +((ub4)k[6]<<16) +((ub4)k[7]<<24));
		c += (k[8] +((ub4)k[9]<<8) +((ub4)k[10]<<16)+((ub4)k[11]<<24));
		mix(a,b,c);
		k += 12; len -= 12;
	}

	/*------------------------------------- handle the last 11 bytes */
	c += length;
	switch(len){	/* all the case statements fall through */
	case 11: c+=((ub4)k[10]<<24);
	case 10: c+=((ub4)k[9]<<16);
	case 9 : c+=((ub4)k[8]<<8);
	/* the first byte of c is reserved for the length */
	case 8 : b+=((ub4)k[7]<<24);
	case 7 : b+=((ub4)k[6]<<16);
	case 6 : b+=((ub4)k[5]<<8);
	case 5 : b+=k[4];
	case 4 : a+=((ub4)k[3]<<24);
	case 3 : a+=((ub4)k[2]<<16);
	case 2 : a+=((ub4)k[1]<<8);
	case 1 : a+=k[0];
	/* case 0: nothing left to add */
	}
	mix(a,b,c);
	/*-------------------------------------------- report the result */
	return c;
}

size_t
__gnu_cxx::hash<Node*>::operator()(Node *nd) const
{
	size_t r;
	ub1 *hk;

	assert(nd != NULL);
	hk = (ub1 *) ((nd->GetPackedState())->hash_key);
	r = jenkins_hash(hk, StateHash::HASH_KEY_BYTE_SIZE, 0x9e3779b9);
	return r;
}

bool
eqnode::operator()(Node *n1, Node *n2) const
{
	bool r;

	assert(n1 != NULL);
	assert(n2 != NULL);
	r = *n1 == *n2;
	return r;
}
