class Node;
typedef struct NodeElement NodeElement;
typedef struct PackedState PackedState;

namespace __gnu_cxx
{
	template<> struct hash<Node *>{
		size_t operator()(Node *) const;
	};
}
struct eqnode{
	bool operator()(Node*, Node*) const;
};
typedef __gnu_cxx::hash_set<Node *, __gnu_cxx::hash<Node *>, eqnode> NodeHashSet;

struct NodeElement{
	Node	*node;
	NodeElement	*next;
};

class NodeStack{
private:

	NodeElement *top;
	NodeElement *current;
	long	length;
	long	max_length;

public:
	NodeStack();
	~NodeStack();

	bool Empty();
	Node *Top();

	void Push(Node *node);
	Node *Pop();
	Node *OnStack(Node *node);

	void ResetIterator();
	Node* Next();	/* May return NULL. */

	void Print();

	long GetUsedMem();
	long GetLength();
};

struct PackedState{
	uchar	*hash_key;
	int	ltl_trans_id;
	int	autom_trans_id;
	short	autom_number;
};

class Node{
private:

	StateBase *state;
	PackedState *packed_state;
	Node *parent;
	uchar m_flags;
	int lowlink;
	int number;
	int component_number;
	NodeElement *Children;	/* list of children */
	NodeElement *LastChild;	/* pointer into Children */
	
	friend class Graph;

public:

	enum NodeFlags{
		NONE	= 0,
		FORCED_ACCEPTED	= 1,
		FORCED_FINAL	= 2,
		IN_SCENARIO	= 4,
		R_SCENARIO	= 8,
		/*I_ACCEPTANCE	= 16,*/
		FIXED	= 32,
		PASSED	= 64,
		TRANS_TO_BORDER	= 128,
	};

	Node(StateBase *st, short autom_num, int autom_trans_id, int ltl_trans_id);
	~Node();

	bool operator==(Node &other);

	/*
		PDS: lowlink, number, state, packed_state, parent,
		flags, and component_number should simply be public.
	*/
	
	void SetLowLink(int);
	void SetNumber(int);
	int  GetLowLink();
	int  GetNumber();

	StateBase *GetState();
	PackedState *GetPackedState();

	Node* GetParent();

	/*
		Hideous:
		
		GetNextChild returns the node's children in turn.
		GetLastChild returns the child most recently returned
		by GetNextChild and GetCurrNodeElement returns its
		enclosing NodeElement.  That element can be used with
		SetCurrNodeElement to go back to that child.
		ResetLastChild goes back to the first child.
		LastChildIsFirst returns true at the first child.
	*/
	void ResetLastChild();
	Node *GetNextChild();
	Node* GetLastChild();
	bool LastChildIsFirst();
	void SetCurrNodeElement(NodeElement*);
	NodeElement *GetCurrNodeElement();

	int GetComponentNumber();
	void SetComponentNumber(int);

	/*
		Flags
	*/
	bool InScenario();	/* IN_SCENARIO */
	bool InScenarioNotInRS();	/* R_SCENARIO */
	bool IsFixed();	/* FIXED */
	bool IsPassed();	/* PASSED */
	bool HasBorderChild();	/* TRANS_TO_BORDER */
	void SetFlags(uchar flags, bool enable);
	void SetInScenario(bool);	/* IN_SCENARIO */
	void SetInScenarioNotInRS(bool);	/* R_SCENARIO */
	void SetFixed(bool);	/* FIXED */
	void SetPassed(bool);	/* PASSED */
	void SetBorderTransStart(bool);	/* TRANS_TO_BORDER */
	void ForceAccepting(bool);	/* Set or clear FORCED_ACCEPTED. */
	void ForceFinal(bool);	/* Set or clear FORCED_FINAL. */

	/*
		GenerateNodeState helper.
		If procnum<0, works on the never claim.

		PDS: It should just return the list.
	*/
private:
	void GenerateAtomicStates(int procnum, StateBase *state, list<StateBase*> &states);
public:
	StateBase* GenerateNodeState();	/* Reconstruct packed state */
	
	/*
		GenerateChildren helpers.
	*/
private:
	void AutomsProcesing(StateBase *state, StateBase *NextState_never, int ltl_trans_id);
	void AddChild(Node* node);
	Node* GenerateChild(int procnum, int pr_trans_id, int trans_id);

  void GenerateChild(StateBase *NextState, short procnum, int autom_trans_id, int ltl_trans_id);

public:
	void GenerateChildren();

	bool HasSelfLoop();	/* Ie, an arrow this -> this */
	bool Unexplored();

	bool IsAccepting();
	bool IsFinal();

	void Print();

private:
	void PrintSubgraphDotFormat(FILE *fd);

public:
	void PackNode();
	unsigned int CalcSize(bool involve_states = false);

	void LeftyInsertNode(FILE *fd, char *GraphTableName, bool DFS_output = false);
	void LeftyInsertEdge(FILE *fd, Node *source, Node *sink, char *GraphTableName, bool insert_to_file = true, bool DFS_output = false);
	void LeftyUpdateNode(FILE *fd, char *GraphTableName);
	void LeftyInsertNodeStep(FILE *fd, char *GraphTableName);
	void LeftyInsertEdgeStep(FILE *fd, Node *source, Node *sink, char *GraphTableName);
	void LeftyUpdateNodeStep(FILE *fd, char *GraphTableName);
	void LeftyHighlight(FILE *fd, char *GraphTableName);
	void LeftyHighlightLabel(FILE *fd, char *GraphTableName);
	void LeftyHighlightEdge(FILE *fd, Node *source, Node *sink, char *GraphTableName);
	void PrintLeftyProgram(FILE *fd, char *GraphTableName);

public:
	void ProfileHashCollisions();	/* PDS: Dead code. */

	void RemoveDuplicatedChildren();
};

struct GraphMetrics{
	int acceptance_number;
	int max_node_out_degree;
	int max_not_final_out_degree;
	int graph_depth;
	int longest_scenario;
	int shortest_scenario;

	GraphMetrics();
};

class Graph{
private:

	list<Node*> GraphNodes;
  
public:

	unsigned int edgecount;

public:

	Graph();

	void Add(Node*);
	void Add(NodeStack &nodes);
	
	void PrintToFile();
	
	unsigned int GetNodeCount();
	unsigned int CalcSize();
	
	GraphMetrics CountGraphMetrics(Node *initial);
};

extern ofstream	graph_file;
extern ofstream	dfs_graph_file;
extern bool	ascii_file;
extern bool	bin32_file;
extern bool	do_lefty_output;
extern bool	supertrace;
extern bool	detect_collisions;
extern long	dfs_edges_count;
extern long	hash_comparison;
extern long	state_restoring;
extern long	state_comparison;
extern long	collisions_count;
extern bool	edge_capping;
extern unsigned long	reconstructing_stack_summ;
extern long	max_depth;
extern NodeHashSet	Nodes;
extern list<Node*>	NodeList;
